import datetime
import logging
import typing

from django.core.cache import cache
from django.core.exceptions import MiddlewareNotUsed
from django.http import HttpRequest, HttpResponse
from django.utils import timezone

from .models import RequestParser, UserVisit
from .settings import RECORDING_DISABLED

logger = logging.getLogger(__name__)


def seconds_to_midnight() -> int:
    """Calculate seconds from now to midnight."""
    now = timezone.now()
    tomorrow = now + datetime.timedelta(days=1)
    midnight = datetime.datetime.combine(tomorrow, datetime.time.min, tzinfo=now.tzinfo)
    return (midnight - now).seconds


def check_cache(parser: RequestParser) -> bool:
    """Check cache for parser and return True if the hashes match."""
    return cache.get(parser.cache_key) == hash(parser)


def update_cache(parser: RequestParser) -> None:
    """Cache instance of a RequestParser."""
    cache.set(parser.cache_key, hash(parser), timeout=seconds_to_midnight())


class UserVisitMiddleware:
    """
    Middleware to record user visits.

    This middleware caches visits on a daily basis. Same user, device, IP
    and session - no new visit.

    """

    def __init__(self, get_response: typing.Callable) -> None:
        if RECORDING_DISABLED:
            raise MiddlewareNotUsed("UserVisit recording has been disabled")
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> typing.Optional[HttpResponse]:
        # this will fail hard if session or authentication middleware are
        # not configured.
        if request.user.is_anonymous:
            return self.get_response(request)

        parser = RequestParser(request)
        if check_cache(parser):
            return self.get_response(request)

        uv = UserVisit.objects.record(request)
        update_cache(parser)
        logger.debug("Recorded and cached new user visit: %r", uv)

        return self.get_response(request)



from __future__ import annotations

import datetime
import uuid
from typing import Optional

import user_agents
from django.conf import settings
from django.db import models
from django.http import HttpRequest
from django.utils import timezone


class RequestParser:
    """
    Parse HttpRequest object.

    For recording purposes we are only interested in a small subset of the
    full HttpRequest object. This class is used to parse out the properties
    that we need, and also to provide a convenient equivalence mechanic - so
    that we can determine whether to record a request or not. (See the
    __hash__ method for details.)

    """

    def __init__(self, request: HttpRequest) -> None:
        """
        Initialise parser from HttpRequest object.

        Raises ValueError if the request.user is not authenticated.

        """
        if not request.session:
            raise ValueError("Request object has no session.")
        if not request.user:
            raise ValueError("Request object has no user.")
        if request.user.is_anonymous:
            raise ValueError("Request user is anonymous.")
        self.request = request
        # set in constructor as we want this fixed to object creation
        # date, not when someone accesses the object.
        self.date = datetime.date.today()

    def __hash__(self) -> int:
        """
        Return object hash.

        The object hash is used when comparing objects. For this class we want
        object to be considered equal if the request properties we are recording
        are the same, and the object was created on the same *day*.

        """
        return (
            hash(self.user.pk)
            ^ hash(self.date)
            ^ hash(self.remote_addr)
            ^ hash(self.session_key)
            ^ hash(self.ua_string)
        )

    @property
    def user(self) -> settings.AUTH_USER_MODEL:
        return self.request.user

    @property
    def remote_addr(self) -> str:
        """Extract client IP from request."""
        x_forwarded_for = self.request.headers.get("X-Forwarded-For")
        if x_forwarded_for:
            return x_forwarded_for.split(",")[0]
        return self.request.META.get("REMOTE_ADDR", "")

    @property
    def session_key(self) -> str:
        """Extract session id from request."""
        return self.request.session.session_key or ""

    @property
    def ua_string(self) -> str:
        """Extract client user-agent from request."""
        return self.request.headers.get("User-Agent", "")

    @property
    def cache_key(self) -> str:
        """Return key used for caching object."""
        return f"user_visit:{self.user.pk}"


class UserVisitManager(models.Manager):
    """Custom model manager for UserVisit objects."""

    def record(
        self, request: HttpRequest, timestamp: datetime.datetime = timezone.now()
    ) -> Optional[UserVisit]:
        """
        Record a new user visit.

        This method will look for an existing UserVisit for the date (extracted
        from the timestamp), matching the session, user-agent and remote_addr
        properties. If any of these have changed, we create a new object. This
        ensures that we get one record per day, per user, per session, device
        and IP address. If any of these change, we get a new record.

        Returns the UserVisit object that is found or created.

        """
        parser = RequestParser(request)
        try:
            uv = UserVisit.objects.get(
                user=request.user,
                timestamp__date=timestamp.date(),
                session_key=parser.session_key,
                ua_string=parser.ua_string,
                remote_addr=parser.remote_addr,
            )
        except UserVisit.DoesNotExist:
            uv = UserVisit.objects.create(
                user=request.user,
                timestamp=timestamp,
                session_key=parser.session_key,
                ua_string=parser.ua_string,
                remote_addr=parser.remote_addr,
            )
        # this should never happen, but race condition.
        except UserVisit.MultipleObjectsReturned:
            uv = UserVisit.objects.filter(
                user=request.user,
                timestamp__date=timestamp.date(),
                session_key=parser.session_key,
                ua_string=parser.ua_string,
                remote_addr=parser.remote_addr,
            ).last()
        return uv


class UserVisit(models.Model):
    """
    Record of a user visiting the site on a given day.

    This is used for tracking and reporting - knowing the volume of visitors
    to the site, and being able to report on someone's interaction with the site.

    We record minimal info required to identify user sessions, plus changes in
    IP and device. This is useful in identifying suspicious activity (multiple
    logins from different locations).

    Also helpful in identifying support issues (as getting useful browser data
    out of users can be very difficult over live chat).

    """

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name="user_visits", on_delete=models.CASCADE
    )
    timestamp = models.DateTimeField(
        help_text="The time at which the first visit of the day was recorded",
        default=timezone.now,
    )
    session_key = models.CharField(help_text="Django session identifier", max_length=40)
    remote_addr = models.CharField(
        help_text=(
            "Client IP address (from X-Forwarded-For HTTP header, "
            "or REMOTE_ADDR request property)"
        ),
        max_length=100,
        blank=True,
    )
    ua_string = models.TextField(
        "User agent (raw)", help_text="Client User-Agent HTTP header", blank=True,
    )
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)

    objects = UserVisitManager()

    def __str__(self) -> str:
        return f"{self.user} visited the site on {self.timestamp}"

    def __repr__(self) -> str:
        return f"<UserVisit user_id={self.user_id} date='{self.date}'>"

    @property
    def user_agent(self) -> user_agents.parsers.UserAgent:
        """Return UserAgent object from the raw user_agent string."""
        return user_agents.parsers.parse(self.ua_string)

    @property
    def date(self) -> datetime.date:
        """Extract the date of the visit from the timestamp."""
        return self.timestamp.date()



# Generated by Django 3.0.8 on 2020-07-02 20:27

import uuid

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="UserVisit",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "timestamp",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        help_text=(
                            "The time at which the first visit of the day was recorded"
                        ),
                    ),
                ),
                (
                    "session_key",
                    models.CharField(
                        help_text="Django session identifier", max_length=40
                    ),
                ),
                (
                    "remote_addr",
                    models.CharField(
                        blank=True,
                        help_text=(
                            "Client IP address (from X-Forwarded-For HTTP header, "
                            "or REMOTE_ADDR request property)"
                        ),
                        max_length=100,
                    ),
                ),
                (
                    "ua_string",
                    models.TextField(
                        blank=True,
                        help_text="Client User-Agent HTTP header",
                        verbose_name="User agent (raw)",
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="user_visits",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



from os import getenv
from typing import Any, Callable

from django.conf import settings


def _env_or_setting(key: str, default: Any, cast_func: Callable = lambda x: x) -> Any:
    return cast_func(getenv(key) or getattr(settings, key, default))


RECORDING_DISABLED = _env_or_setting(
    "USER_VISIT_RECORDING_DISABLED", False, lambda x: bool(x)
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['user_visit', 'user_visit.migrations']

package_data = \
{'': ['*']}

install_requires = \
['django>=2.2,<4.0', 'user-agents>=2.1,<3.0']

setup_kwargs = {
    'name': 'django-user-visit',
    'version': '0.1',
    'description': 'Django app used to track user visits.',
    'long_description': "# django-user-visit\n\nDjango app for recording daily user visits\n\nThis app consists of middleware to record user visits, and a\nsingle `UserVisit` model to capture that data.\n\nThe principal behind this is _not_ to record every single\nrequest made by a user. It is to record each daily visit to\na site.\n\nThe one additional factor is that it will record a single daily\nvisit per user/session/device/ip combination. This means that if\na user visits a site multiple times from the same location / same\ndevice, without logging out, then they will be recorded once. If\nthe same user logs in from a different device, IP address, then\nthey will be recorded again.\n\nThe goal is to record unique daily visits per user 'context' (\nwhere context is the location / device combo).\n\nAdmin list view:\n\n![UserVisit list view](assets/screenshot-admin-list-view.png)\n\nAdmin edit view:\n\n![UserVisit edit view](assets/screenshot-admin-edit-view.png)\n",
    'author': 'YunoJuno',
    'author_email': 'code@yunojuno.com',
    'maintainer': 'YunoJuno',
    'maintainer_email': 'code@yunojuno.com',
    'url': 'https://github.com/yunojuno/django-user-visit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from django.contrib import admin

from .models import UserVisit


class UserVisitAdmin(admin.ModelAdmin):

    list_display = ("timestamp", "user", "session_key", "remote_addr", "user_agent")
    list_filter = ("timestamp",)
    search_fields = (
        "user__first_name",
        "user__last_name",
        "user__username",
        "ua_string",
    )
    raw_id_fields = ("user",)
    readonly_fields = (
        "user",
        "timestamp",
        "session_key",
        "remote_addr",
        "user_agent",
        "ua_string",
    )
    ordering = ("-timestamp",)


admin.site.register(UserVisit, UserVisitAdmin)
